@@ -31,12 +31,24 @@ CallbackType = Union[Callable[[], Awaitable[None]], Callable[[], None]]
 """Type used for arbitrary callables used in callbacks."""
 IgnoreReturnCallbackType = Union[Callable[[], Awaitable[Any]], Callable[[], Any]]
 """A callback which ignores the return type."""
-WatchCallbackType = Union[
-    Callable[[], Awaitable[None]],
-    Callable[[Any], Awaitable[None]],
+WatchCallbackBothValuesType = Union[
     Callable[[Any, Any], Awaitable[None]],
-    Callable[[], None],
-    Callable[[Any], None],
     Callable[[Any, Any], None],
 ]
+"""Type for watch methods that accept the old and new values of reactive objects."""
+WatchCallbackNewValueType = Union[
+    Callable[[Any], Awaitable[None]],
+    Callable[[Any], None],
+]
+"""Type for watch methods that accept only the new value of reactive objects."""
+WatchCallbackNoArgsType = Union[
+    Callable[[], Awaitable[None]],
+    Callable[[], None],
+]
+"""Type for watch methods that do not require the explicit value of the reactive."""
+WatchCallbackType = Union[
+    WatchCallbackBothValuesType,
+    WatchCallbackNewValueType,
+    WatchCallbackNoArgsType,
+]
 """Type used for callbacks passed to the `watch` method of widgets."""
